from typing import Any
from dataclasses import dataclass
from functools import cached_property

from trajax.types import (
    StateSequence,
    SimulationData,
    Metric,
    Positions,
    PositionExtractor,
    Trajectory,
)
from trajax.collectors import access

from numtypes import BoolArray, Array, Dims, D, array

import numpy as np


@dataclass(kw_only=True, frozen=True)
class TaskCompletionMetricResult[T: int = int]:
    completion: BoolArray[Dims[T]]
    completion_time: float
    efficiency: float

    @cached_property
    def completed(self) -> bool:
        return bool(self.completion.any())


@dataclass(kw_only=True, frozen=True)
class TaskCompletionMetric[StateBatchT](Metric[TaskCompletionMetricResult[Any]]):
    goal_position: Array[Dims[D[2]]]
    optimal_distance: float
    distance_threshold: float
    time_step_size: float
    position_extractor: PositionExtractor[StateBatchT, Positions]

    @staticmethod
    def create(
        *,
        reference: Trajectory[Any, Any],
        distance_threshold: float,
        time_step_size: float,
        position_extractor: PositionExtractor[StateBatchT, Positions],
    ) -> "TaskCompletionMetric[StateBatchT]":
        assert reference.natural_length > 0.0, (
            f"Reference trajectory must have positive length, got {reference.natural_length}."
        )

        return TaskCompletionMetric(
            goal_position=array(reference.end, shape=(2,)),
            optimal_distance=reference.natural_length,
            distance_threshold=distance_threshold,
            time_step_size=time_step_size,
            position_extractor=position_extractor,
        )

    def compute[T: int = int](
        self, data: SimulationData
    ) -> TaskCompletionMetricResult[T]:
        states = data(
            access.states.assume(StateSequence[T, Any, StateBatchT]).require()
        )
        positions = np.asarray(self.position_extractor(states.batched()))[..., 0]

        distances_from_goal = np.linalg.norm(positions - self.goal_position, axis=1)
        completion = distances_from_goal <= self.distance_threshold
        traversed_distance = np.linalg.norm(np.diff(positions, axis=0), axis=1).sum()

        return TaskCompletionMetricResult(
            completion=completion,
            completion_time=(
                float(np.argmax(completion)) * self.time_step_size
                if completion.any()
                else float("inf")
            ),
            efficiency=traversed_distance / self.optimal_distance,
        )

    @property
    def name(self) -> str:
        return "task-completion"



from .common import (
    BoundaryDistance as BoundaryDistance,
    BoundaryDistanceExtractor as BoundaryDistanceExtractor,
)
from .basic import (
    NumPyBoundaryDistance as NumPyBoundaryDistance,
    NumPyBoundaryDistanceExtractor as NumPyBoundaryDistanceExtractor,
)
from .accelerated import (
    JaxBoundaryDistance as JaxBoundaryDistance,
    JaxBoundaryDistanceExtractor as JaxBoundaryDistanceExtractor,
)



from typing import Any
from dataclasses import dataclass

from trajax.types import (
    CostFunction,
    ControlInputBatch,
    Trajectory,
    NumPyBoundaryDistance,
    NumPyBoundaryDistanceExtractor,
    NumPyCosts,
    NumPyPositions,
    NumPyLateralPositions,
    NumPyPositionExtractor,
)
from trajax.states import NumPySimpleCosts

import numpy as np


@dataclass(frozen=True)
class NumPyBoundaryCost[StateT](CostFunction[ControlInputBatch, StateT, NumPyCosts]):
    distance: NumPyBoundaryDistanceExtractor[StateT, NumPyBoundaryDistance]
    distance_threshold: float
    weight: float

    @staticmethod
    def create[S](
        *,
        distance: NumPyBoundaryDistanceExtractor[S, NumPyBoundaryDistance],
        distance_threshold: float,
        weight: float,
    ) -> "NumPyBoundaryCost[S]":
        return NumPyBoundaryCost(
            distance=distance, distance_threshold=distance_threshold, weight=weight
        )

    def __call__[T: int, M: int](
        self, *, inputs: ControlInputBatch[T, int, M], states: StateT
    ) -> NumPyCosts[T, M]:
        cost = self.distance_threshold - self.distance(states=states).array

        return NumPySimpleCosts(self.weight * np.clip(cost, 0, None))


@dataclass(kw_only=True, frozen=True)
class NumPyFixedWidthBoundary[StateT](
    NumPyBoundaryDistanceExtractor[StateT, NumPyBoundaryDistance]
):
    reference: Trajectory[Any, Any, NumPyPositions, NumPyLateralPositions]
    position_extractor: NumPyPositionExtractor[StateT]
    left: float
    right: float

    @staticmethod
    def create[S](
        *,
        reference: Trajectory[Any, Any, NumPyPositions, NumPyLateralPositions],
        position_extractor: NumPyPositionExtractor[S],
        left: float,
        right: float,
    ) -> "NumPyFixedWidthBoundary[S]":
        """Creates a fixed-width boundary distance extractor.

        This component assumes a fixed-width corridor around a reference trajectory. The left and
        right widths can be different (asymmetric corridor).

        Args:
            reference: The reference trajectory defining the center of the corridor.
            position_extractor: Function to extract positions from states.
            left: The width of the left side of the corridor.
            right: The width of the right side of the corridor.
        """
        assert left >= -right, (
            f"The boundaries appear to be inverted. Left: {left}, Right: {right}. "
            f"Make sure the total width (left + right) is non-negative, got {left + right}."
        )

        return NumPyFixedWidthBoundary(
            reference=reference,
            position_extractor=position_extractor,
            left=left,
            right=right,
        )

    def __call__(self, *, states: StateT) -> NumPyBoundaryDistance:
        positions = self.position_extractor(states)
        lateral = self.reference.lateral(positions)

        distance_to_left = self.left + lateral.array
        distance_to_right = self.right - lateral.array

        return NumPyBoundaryDistance(np.minimum(distance_to_left, distance_to_right))



from typing import Protocol

from trajax.types.array import DataType

from numtypes import Array, Dims


class BoundaryDistance[T: int, M: int](Protocol):
    def __array__(self, dtype: DataType | None = None) -> Array[Dims[T, M]]:
        """Returns the distances between the ego and the nearest boundary as a NumPy array."""
        ...

    @property
    def horizon(self) -> T:
        """The time horizon over which the distances are defined."""
        ...

    @property
    def rollout_count(self) -> M:
        """The number of rollouts for which the distances are defined."""
        ...


class BoundaryDistanceExtractor[StateBatchT, DistanceT](Protocol):
    def __call__(self, *, states: StateBatchT) -> DistanceT:
        """Computes the distances between the ego and the nearest boundary."""
        ...



from typing import Any
from dataclasses import dataclass

from trajax.types import (
    jaxtyped,
    CostFunction,
    ControlInputBatch,
    Trajectory,
    JaxBoundaryDistance,
    JaxBoundaryDistanceExtractor,
    JaxCosts,
    JaxPositions,
    JaxLateralPositions,
    JaxPositionExtractor,
)
from trajax.states import JaxSimpleCosts

from jaxtyping import Float, Array as JaxArray, Scalar

import jax
import jax.numpy as jnp


@dataclass(frozen=True)
class JaxBoundaryCost[StateT](CostFunction[ControlInputBatch, StateT, JaxCosts]):
    distance: JaxBoundaryDistanceExtractor[StateT, JaxBoundaryDistance]
    distance_threshold: Scalar
    weight: Scalar

    @staticmethod
    def create[S](
        *,
        distance: JaxBoundaryDistanceExtractor[S, JaxBoundaryDistance],
        distance_threshold: float,
        weight: float,
    ) -> "JaxBoundaryCost[S]":
        return JaxBoundaryCost(
            distance=distance,
            distance_threshold=jnp.array(distance_threshold),
            weight=jnp.array(weight),
        )

    def __call__[T: int, M: int](
        self, *, inputs: ControlInputBatch[T, int, M], states: StateT
    ) -> JaxCosts[T, M]:
        return JaxSimpleCosts(
            boundary_cost(
                distance=self.distance(states=states).array,
                distance_threshold=self.distance_threshold,
                weight=self.weight,
            )
        )


@dataclass(kw_only=True, frozen=True)
class JaxFixedWidthBoundary[StateT](
    JaxBoundaryDistanceExtractor[StateT, JaxBoundaryDistance]
):
    reference: Trajectory[Any, Any, JaxPositions, JaxLateralPositions]
    position_extractor: JaxPositionExtractor[StateT]
    left: Scalar
    right: Scalar

    @staticmethod
    def create[S](
        *,
        reference: Trajectory[Any, Any, JaxPositions, JaxLateralPositions],
        position_extractor: JaxPositionExtractor[S],
        left: float,
        right: float,
    ) -> "JaxFixedWidthBoundary[S]":
        """Creates a fixed-width boundary distance extractor.

        This component assumes a fixed-width corridor around a reference trajectory. The left and
        right widths can be different (asymmetric corridor).

        Args:
            reference: The reference trajectory defining the center of the corridor.
            position_extractor: Function to extract positions from states.
            left: The width of the left side of the corridor.
            right: The width of the right side of the corridor.
        """
        assert left >= -right, (
            f"The boundaries appear to be inverted. Left: {left}, Right: {right}. "
            f"Make sure the total width (left + right) is non-negative, got {left + right}."
        )

        return JaxFixedWidthBoundary(
            reference=reference,
            position_extractor=position_extractor,
            left=jnp.array(left),
            right=jnp.array(right),
        )

    def __call__(self, *, states: StateT) -> JaxBoundaryDistance:
        positions = self.position_extractor(states)
        lateral = self.reference.lateral(positions)

        return JaxBoundaryDistance(
            boundary_distance(lateral=lateral.array, left=self.left, right=self.right)
        )


@jax.jit
@jaxtyped
def boundary_cost(
    *, distance: Float[JaxArray, "T M"], distance_threshold: Scalar, weight: Scalar
) -> Float[JaxArray, "T M"]:
    cost = distance_threshold - distance
    return weight * jnp.clip(cost, 0, None)


@jax.jit
@jaxtyped
def boundary_distance(
    *, lateral: Float[JaxArray, "T M"], left: Scalar, right: Scalar
) -> Float[JaxArray, "T M"]:
    distance_to_left = left + lateral
    distance_to_right = right - lateral
    return jnp.minimum(distance_to_left, distance_to_right)
